"""Authentication Models."""

from dataclasses import dataclass, field
import datetime as dt

from spnkr.models import SnakeModel, Date, PascalModel
from spnkr import util


@dataclass
class XTokenResponse(PascalModel):
    issue_instant: dt.datetime
    not_after: dt.datetime
    token: str

    def is_valid(self) -> bool:
        return self.not_after > util.utc_now()


@dataclass
class XAUDisplayClaims(SnakeModel):
    xui: list[dict[str, str]]


@dataclass
class XAUResponse(XTokenResponse):
    display_claims: XAUDisplayClaims


@dataclass
class XSTSDisplayClaims(SnakeModel):
    xui: list[dict[str, str]]


@dataclass
class XSTSResponse(XTokenResponse):
    display_claims: XSTSDisplayClaims

    @property
    def xuid(self) -> str:
        return self.display_claims.xui[0]["xid"]

    @property
    def userhash(self) -> str:
        return self.display_claims.xui[0]["uhs"]

    @property
    def gamertag(self) -> str:
        return self.display_claims.xui[0]["gtg"]

    @property
    def age_group(self) -> str:
        return self.display_claims.xui[0]["agg"]

    @property
    def privileges(self) -> str:
        return self.display_claims.xui[0]["prv"]

    @property
    def user_privileges(self) -> str:
        return self.display_claims.xui[0]["usr"]

    @property
    def authorization_header_value(self) -> str:
        return f"XBL3.0 x={self.userhash};{self.token}"


@dataclass
class OAuth2TokenResponse(SnakeModel):
    token_type: str
    expires_in: int
    scope: str
    access_token: str
    refresh_token: str | None
    user_id: str
    issued: dt.datetime = field(default_factory=util.utc_now)

    def is_valid(self) -> bool:
        return (self.issued + dt.timedelta(seconds=self.expires_in)) > util.utc_now()


@dataclass
class SpartanTokenResponse(PascalModel):
    expires_utc: Date
    spartan_token: str
    token_duration: dt.timedelta

    def is_valid(self) -> bool:
        return self.expires_utc.iso_8601_date > util.utc_now()


@dataclass
class ClearanceTokenResponse(PascalModel):
    flight_configuration_id: str
    



"""Endpoints for the "HIUGC_Discovery" and "HIUGC_Discovery_Open" authorities."""

from spnkr.api.enums import AssetKind, AuthenticationMethod, ResultOrder, SearchProperty
from spnkr.api.authorities.base import BaseAuthority
from spnkr.api.authorities.ugc_discovery import models


class UgcDiscoveryAuthority(BaseAuthority):
    URL = 'https://discovery-infiniteugc.svc.halowaypoint.com:443'

    def get_manifest(self,
                     asset_id: str,
                     version_id: str,
                     clearance_id: str | None = None) -> models.Manifest:
        url = self.URL + f'/hi/manifests/{asset_id}/versions/{version_id}'
        params = dict(clearanceId=clearance_id) if clearance_id is not None else None
        resp = self._session.get(url, params=params)
        resp.raise_for_status()
        return models.Manifest.parse_json(resp.text)

    def get_manifest_by_build(self, build_number: str) -> models.Manifest:
        url = self.URL + f'/hi/manifests/builds/{build_number}/game'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Manifest.parse_json(resp.text)

    def get_custom_game_manifest(self) -> models.Manifest:
        url = self.URL + '/hi/projects/a9dc0785-2a99-4fec-ba6e-0216feaaf041'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Manifest.parse_json(resp.text)

    def get_engine_gamevariant(self, asset_id: str, version_id: str) -> models.EngineGameVariant:
        url = self.URL + f'/hi/engineGameVariants/{asset_id}/versions/{version_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.EngineGameVariant.parse_json(resp.text)

    def get_engine_game_variant_without_version(self, asset_id: str) -> models.EngineGameVariant:
        url = self.URL + f'/hi/engineGameVariants/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.EngineGameVariant.parse_json(resp.text)

    def get_ugc_game_variant(self, asset_id: str, version_id: str) -> models.UgcGameVariant:
        url = self.URL + f'/hi/ugcGameVariants/{asset_id}/versions/{version_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.UgcGameVariant.parse_json(resp.text)

    def get_ugc_game_variant_without_version(self, asset_id: str) -> models.UgcGameVariant:
        url = self.URL + f'/hi/ugcGameVariants/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.UgcGameVariant.parse_json(resp.text)

    def get_map_mode_pair(self,
                          asset_id: str,
                          version_id: str,
                          clearance_id: str | None = None) -> models.MapModePair:
        url = self.URL + f'/hi/mapModePairs/{asset_id}/versions/{version_id}'
        params = dict(clearanceId=clearance_id) if clearance_id is not None else None
        resp = self._session.get(url, params=params)
        resp.raise_for_status()
        return models.MapModePair.parse_json(resp.text)

    def get_map_mode_pair_without_version(self, asset_id: str) -> models.MapModePair:
        """Work in progress, currently only returns 404."""
        url = self.URL + f'/hi/mapModePairs/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return resp.text  # type: ignore

    def get_map(self, asset_id: str, version_id: str) -> models.Map:
        url = self.URL + f'/hi/maps/{asset_id}/versions/{version_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Map.parse_json(resp.text)

    def get_map_without_version(self, asset_id: str) -> models.Map:
        url = self.URL + f'/hi/maps/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Map.parse_json(resp.text)

    def get_playlist(self,
                     asset_id: str,
                     version_id: str,
                     clearance_id: str | None = None) -> models.Playlist:
        url = self.URL + f'/hi/playlists/{asset_id}/versions/{version_id}'
        params = dict(clearanceId=clearance_id) if clearance_id is not None else None
        resp = self._session.get(url, params=params)
        resp.raise_for_status()
        return models.Playlist.parse_json(resp.text)

    def get_playlist_without_version(self, asset_id: str) -> models.Playlist:
        url = self.URL + f'/hi/playlists/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Playlist.parse_json(resp.text)

    def get_prefab(self, asset_id: str, version_id: str) -> models.Prefab:
        url = self.URL + f'/hi/prefabs/{asset_id}/versions/{version_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Prefab.parse_json(resp.text)

    def get_prefab_without_version(self, asset_id: str) -> models.Prefab:
        url = self.URL + f'/hi/prefabs/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Prefab.parse_json(resp.text)

    def get_project(self, asset_id: str, version_id: str) -> models.Project:
        url = self.URL + f'/hi/projects/{asset_id}/versions/{version_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Project.parse_json(resp.text)

    def get_project_without_version(self, asset_id: str) -> models.Project:
        url = self.URL + f'/hi/projects/{asset_id}'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Project.parse_json(resp.text)

    def get_tags_info(self) -> models.TagsInfo:
        url = self.URL + '/hi/info/tags'
        resp = self._session.get(url, auth_method=AuthenticationMethod.ClearanceToken)
        resp.raise_for_status()
        return models.TagsInfo.parse_json(resp.text)

    def search(self,
               start: int = 0,
               count: int = 101,
               sort: SearchProperty = SearchProperty.PlaysRecent,
               order: ResultOrder = ResultOrder.Desc,
               asset_kind: AssetKind | None = None) -> models.Search:
        """Search for assets in the user-generated content directory.

        Args:
            start (int): Number of results from which to offset the result set, starting at 0.
            count (int): Count of results to return. Max 101.
            sort (enums.SearchProperty): Property by which to sort the results. Example is "PlaysRecent".
            order (enums.ResultOrder): Determines whether results are ordered in descending or ascending order.
            asset_kind (enums.AssetKind): Type of asset to be searched.
        """
        url = self.URL + '/hi/search'
        params = {
            'start': start,
            'count': count,
            'sort': sort.name,
            'order': order.name,
            'assetKind': asset_kind.name if asset_kind is not None else None
        }
        resp = self._session.get(url, params=params)
        resp.raise_for_status()
        return models.Search.parse_json(resp.text)

    def spectate_by_match_id(self, match_id: str) -> models.Film:
        url = self.URL + f'/hi/films/matches/{match_id}/spectate'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.Film.parse_json(resp.text)

    # HTTP 404 - Not Found for "6369c3a6-390e-496c-ab71-93c326347327" asset id on 9/9/2022
    # def get_manifest_without_version(self, asset_id: str):
    #     url = self.URL + f'/hi/manifests/{asset_id}'
    #     resp = self._session.get(url)
    #     resp.raise_for_status()
    #     return resp.text

    # HTTP 403 - Forbidden for "HIRC" branch name on 9/9/2022
    # def get_manifest_by_branch(self, branch_name: str):
    #     url = self.URL + f'/hi/manifests/branches/{branch_name}/game'
    #     resp = self._session.get(url)
    #     resp.raise_for_status()
    #     return resp.text




"""
Authentication Manager

Authenticate with Windows Live Server and Xbox Live.
"""

import urllib.parse

import requests

from spnkr.authentication import models as mdl
from spnkr.exceptions import AuthenticationException

DEFAULT_SCOPES = ['Xboxlive.signin', 'Xboxlive.offline_access']
XSTS_V3_XBOX_AUDIENCE = 'http://xboxlive.com'
XSTS_V3_HALO_AUDIENCE = 'https://prod.xsts.halowaypoint.com/'


class AuthenticationManager:
    def __init__(self,
                 client_session: requests.Session,
                 client_id: str,
                 client_secret: str,
                 redirect_uri: str,
                 scopes: list[str] | None = None):
        self.session = client_session
        self._client_id: str = client_id
        self._client_secret: str = client_secret
        self._redirect_uri: str = redirect_uri
        self._scopes: list[str] = scopes or DEFAULT_SCOPES

        self.oauth: mdl.OAuth2TokenResponse | None = None
        self.user_token: mdl.XAUResponse | None = None
        self.xsts_token: mdl.XSTSResponse | None = None
        self.halo_xsts_token: mdl.XSTSResponse | None = None
        self.spartan_token: mdl.SpartanTokenResponse | None = None
        self.clearance_token: mdl.ClearanceTokenResponse | None = None

    def generate_authorization_url(self, state: str | None = None) -> str:
        """Generate Windows Live Authorization URL."""
        params = {
            'client_id': self._client_id,
            'response_type': 'code',
            'approval_prompt': 'auto',
            'scope': ' '.join(self._scopes),
            'redirect_uri': self._redirect_uri,
        }

        if state:
            params['state'] = state

        return 'https://login.live.com/oauth20_authorize.srf?' + urllib.parse.urlencode(params)

    def request_tokens(self, authorization_code: str) -> None:
        """Request all tokens."""
        self.oauth = self.request_oauth_token(authorization_code)
        self.user_token = self.request_user_token()
        self.xsts_token = self.request_xsts_token(XSTS_V3_XBOX_AUDIENCE)
        self.halo_xsts_token = self.request_xsts_token(XSTS_V3_HALO_AUDIENCE)
        self.spartan_token = self.request_spartan_token()
        self.clearance_token = self.request_clearance_token()

    def refresh_tokens(self) -> None:
        """Refresh all tokens."""
        if self.oauth is None or not self.oauth.is_valid():
            self.oauth = self.refresh_oauth_token()
        if self.user_token is None or not self.user_token.is_valid():
            self.user_token = self.request_user_token()
        if self.xsts_token is None or not self.xsts_token.is_valid():
            self.xsts_token = self.request_xsts_token(XSTS_V3_XBOX_AUDIENCE)
        if self.halo_xsts_token is None or not self.halo_xsts_token.is_valid():
            self.halo_xsts_token = self.request_xsts_token(XSTS_V3_HALO_AUDIENCE)
        if self.spartan_token is None or not self.spartan_token.is_valid():
            self.spartan_token = self.request_spartan_token()
        self.clearance_token = self.request_clearance_token()

    def request_oauth_token(self, authorization_code: str) -> mdl.OAuth2TokenResponse:
        """Request OAuth2 token."""
        return self._oauth2_token_request({
            'grant_type': 'authorization_code',
            'code': authorization_code,
            'scope': ' '.join(self._scopes),
            'redirect_uri': self._redirect_uri,
        })

    def refresh_oauth_token(self) -> mdl.OAuth2TokenResponse:
        """Refresh OAuth2 token."""
        return self._oauth2_token_request({
            'grant_type': 'refresh_token',
            'scope': ' '.join(self._scopes),
            'refresh_token': self.oauth.refresh_token,  # type: ignore
        })

    def _oauth2_token_request(self, data: dict[str, str]) -> mdl.OAuth2TokenResponse:
        """Execute token requests."""
        data.update({'client_id': self._client_id, 'client_secret': self._client_secret})
        resp = self.session.post('https://login.live.com/oauth20_token.srf', data=data)
        resp.raise_for_status()
        return mdl.OAuth2TokenResponse.parse_json(resp.text)

    def request_user_token(self,
                           relying_party: str = 'http://auth.xboxlive.com') -> mdl.XAUResponse:
        """Authenticate via access token and receive user token."""
        url = 'https://user.auth.xboxlive.com/user/authenticate'
        headers = {'x-xbl-contract-version': '1'}
        data = {
            'RelyingParty': relying_party,
            'TokenType': 'JWT',
            'Properties': {
                'AuthMethod': 'RPS',
                'SiteName': 'user.auth.xboxlive.com',
                'RpsTicket': f'd={self.oauth.access_token}',  # type: ignore
            },
        }

        resp = self.session.post(url, json=data, headers=headers)
        resp.raise_for_status()
        return mdl.XAUResponse.parse_json(resp.text)

    def request_xsts_token(self, relying_party: str) -> mdl.XSTSResponse:
        """Authorize via user token and receive final X token."""
        url = 'https://xsts.auth.xboxlive.com/xsts/authorize'
        headers = {'x-xbl-contract-version': '1'}
        data = {
            'RelyingParty': relying_party,
            'TokenType': 'JWT',
            'Properties': {
                'UserTokens': [self.user_token.token],  # type: ignore
                'SandboxId': 'RETAIL',
            },
        }

        resp = self.session.post(url, json=data, headers=headers)
        if resp.status_code == 401:
            print('Failed to authorize you! Your password or username may be wrong or you are '
                'trying to use child account (< 18 years old)')
            raise AuthenticationException()
        resp.raise_for_status()
        return mdl.XSTSResponse.parse_json(resp.text)


    def request_spartan_token(self) -> mdl.SpartanTokenResponse:
        url = 'https://settings.svc.halowaypoint.com/spartan-token'
        headers = {'Accept': 'application/json'}
        data = {
            'Audience': 'urn:343:s3:services',
            'MinVersion': '4',
            'Proof': [{
                'Token': self.halo_xsts_token.token,  # type: ignore
                'TokenType': 'Xbox_XSTSv3'
            }]
        }
        resp = self.session.post(url, headers=headers, json=data)
        resp.raise_for_status()
        return mdl.SpartanTokenResponse.parse_json(resp.text)


    def request_clearance_token(self) -> mdl.ClearanceTokenResponse:
        """Request a clearance token, which is required for some endpoints."""
        url = 'https://settings.svc.halowaypoint.com/oban/flight-configurations/titles/hi/audiences/RETAIL/active'
        headers = {
            'x-343-authorization-spartan': self.spartan_token.spartan_token  # type: ignore
        }
        params = {
            'sandbox': 'UNUSED',
            'build': '222249.22.06.08.1730-0'
        }
        resp = self.session.get(url, params=params, headers=headers)
        resp.raise_for_status()
        return mdl.ClearanceTokenResponse.parse_json(resp.text)




""""""

from spnkr.api.authorities.base import BaseAuthority
from spnkr.api.authorities.profile import models
from spnkr.api.enums import AuthenticationMethod, ProfileSetting
from spnkr.exceptions import ApiRateLimitException
from spnkr import util


class ProfileAuthority(BaseAuthority):
    URL = 'https://profile.xboxlive.com:443'
    AUTH_METHOD = AuthenticationMethod.XSTSv3XboxAudience

    def get_profiles(self,
                     xuids: list[str],
                     settings: list[ProfileSetting] | None = None) -> models.ProfileResponse:
        url = self.URL + '/users/batch/profile/settings'
        settings = settings or list(ProfileSetting)
        data = {
            'settings': [ps.value for ps in settings],
            'userIds': [util.unwrap_xuid(x) for x in xuids],
        }
        resp = self._session.post(url, json=data, auth_method=self.AUTH_METHOD)
        if resp.status_code == 429:
            raise ApiRateLimitException('Request rate exceeded')
        resp.raise_for_status()
        return models.ProfileResponse.parse_json(resp.text)

    def get_profile_by_xuid(self,
                            xuid: str,
                            settings: list[ProfileSetting] | None = None
                            ) -> models.ProfileResponse:
        url = self.URL + f'/users/{util.wrap_xuid(xuid)}/profile/settings'
        settings = settings or list(ProfileSetting)
        params = {
            'settings': ','.join([ps.value for ps in settings]),
        }
        resp = self._session.get(url, params=params, auth_method=self.AUTH_METHOD)
        if resp.status_code == 429:
            raise ApiRateLimitException('Request rate exceeded')
        resp.raise_for_status()
        return models.ProfileResponse.parse_json(resp.text)

    def get_profile_by_gamertag(self,
                                gamertag: str,
                                settings: list[ProfileSetting] | None = None
                                ) -> models.ProfileResponse:
        url = self.URL + f'/users/gt({gamertag})/profile/settings'
        settings = settings or list(ProfileSetting)
        params = {
            'settings': ','.join([ps.value for ps in settings]),
        }
        resp = self._session.get(url, params=params, auth_method=self.AUTH_METHOD)
        if resp.status_code == 429:
            raise ApiRateLimitException('Request rate exceeded')
        resp.raise_for_status()
        return models.ProfileResponse.parse_json(resp.text)





from dataclasses import dataclass

from spnkr.models import PascalModel


@dataclass
class CsrContainer(PascalModel):
    value: int
    measurement_matches_remaining: int
    tier: str
    tier_start: int
    sub_tier: int
    next_tier: str
    next_tier_start: int
    next_sub_tier: int
    initial_measurement_matches: int


@dataclass
class StatPerformance(PascalModel):
    count: int
    expected: float
    std_dev: float


@dataclass
class StatPerformances(PascalModel):
    kills: StatPerformance
    deaths: StatPerformance


@dataclass
class RankRecap(PascalModel):
    pre_match_csr: CsrContainer
    post_match_csr: CsrContainer


@dataclass
class Counterfactual(PascalModel):
    kills: float
    deaths: float


@dataclass
class TierCounterfactuals(PascalModel):
    bronze: Counterfactual | None
    silver: Counterfactual | None
    gold: Counterfactual | None
    platinum: Counterfactual | None
    diamond: Counterfactual | None
    onyx: Counterfactual | None


@dataclass
class Counterfactuals(PascalModel):
    self_counterfactuals: Counterfactual
    tier_counterfactuals: TierCounterfactuals


@dataclass
class SkillResult(PascalModel):
    team_mmr: float
    rank_recap: RankRecap
    stat_performances: StatPerformances
    team_id: int
    team_mmrs: dict[int, float]  # keys are strings of team ids (e.g., "0")
    ranked_rewards: None
    counterfactuals: Counterfactuals


@dataclass
class SkillResultContainer(PascalModel):
    id: str
    result_code: int
    result: SkillResult


@dataclass
class MatchSkillInfo(PascalModel):
    value: list[SkillResultContainer]


@dataclass
class PlaylistCsrContainer(PascalModel):
    current: CsrContainer
    season_max: CsrContainer
    all_time_max: CsrContainer


@dataclass
class PlayerPlaylistCsrResult(PascalModel):
    id: str
    result_code: int
    result: PlaylistCsrContainer


@dataclass
class PlaylistCsrInfo(PascalModel):
    value: list[PlayerPlaylistCsrResult]




"""Base Models."""

from dataclasses import dataclass
import datetime as dt
from typing import Type

from dataclass_wizard import JSONSerializable, LoadMixin, json_field
from dataclass_wizard.enums import LetterCase
import dateutil.parser
import orjson

from spnkr import util


def load_to_datetime(o: str, base_type: dt.datetime) -> dt.datetime:
    return dateutil.parser.isoparse(o)

def load_to_timedelta(o: str, base_type: dt.timedelta):
    return util.parse_iso_duration(o)
    
def parse_json(cls: Type[JSONSerializable], string: str):
    return cls.from_json(string, decoder=orjson.loads)  # type: ignore


class SnakeModel(JSONSerializable, LoadMixin):
    load_to_datetime = load_to_datetime
    load_to_timedelta = load_to_timedelta

    @classmethod
    def parse_json(cls, string: str):
        return parse_json(cls, string)


class PascalModel(JSONSerializable, LoadMixin):
    load_to_datetime = load_to_datetime
    load_to_timedelta = load_to_timedelta

    @classmethod
    def parse_json(cls, string: str):
        return parse_json(cls, string)

    class Meta(JSONSerializable.Meta):
        key_transform_with_dump = LetterCase.PASCAL


class CamelModel(JSONSerializable, LoadMixin):
    load_to_datetime = load_to_datetime
    load_to_timedelta = load_to_timedelta

    @classmethod
    def parse_json(cls, string: str):
        return parse_json(cls, string)

    class Meta(JSONSerializable.Meta):
        key_transform_with_dump = LetterCase.CAMEL


@dataclass
class Date(JSONSerializable, LoadMixin):
    iso_8601_date: dt.datetime = json_field(('ISO8601Date',), all=True)

    load_to_datetime = load_to_datetime





import datetime as dt


def unwrap_xuid(xuid: str) -> str:
    return xuid[5:-1] if xuid[0] == 'x' else xuid


def wrap_xuid(xuid: str) -> str:
    return xuid if xuid[0] == 'x' else f'xuid({xuid})'


def parse_iso_duration(iso_duration: str) -> dt.timedelta:
    """Assume element no larger than weeks."""
    assert iso_duration.startswith('PT')
    kwargs = {}
    search = iso_duration[2:]
    seps = {'W': 'weeks', 'D': 'days', 'H': 'hours', 'M': 'minutes', 'S': 'seconds'}
    for sep, attr in seps.items():
        parts = search.split(sep)
        if len(parts) > 1:
            kwargs[attr] = float(parts[0])
            search = parts[1]
    return dt.timedelta(**kwargs)


def utc_now() -> dt.datetime:
    return dt.datetime.now(dt.timezone.utc)

    



from spnkr.api.session import Session
from spnkr.api.authorities import profile, skill, stats, ugc_discovery
from spnkr.authentication.manager import AuthenticationManager


class Client:
    def __init__(self, auth_mgr: AuthenticationManager, validate_tokens: bool = True):
        session = Session(auth_mgr, validate_tokens)

        self.profile = profile.ProfileAuthority(session)
        self.skill = skill.SkillAuthority(session)
        self.stats = stats.StatsAuthority(session)
        self.ugc_discovery = ugc_discovery.UgcDiscoveryAuthority(session)

        



from typing import Any

import ms_cv
import requests

from spnkr.api.enums import AuthenticationMethod
from spnkr.authentication.manager import AuthenticationManager


class Session:
    def __init__(self, auth_mgr: AuthenticationManager, validate_tokens: bool = True):
        self._auth_mgr = auth_mgr
        self._cv = ms_cv.CorrelationVector()
        self._validate_tokens = validate_tokens

    def request(self,
                method: str,
                url: str,
                auth_method: AuthenticationMethod | None = AuthenticationMethod.SpartanToken,
                include_cv: bool = True,
                **kwargs: Any) -> requests.Response:
        """Proxy Request and add Auth/CV headers."""
        headers = kwargs.pop('headers', {})

        if 'Accept' not in headers:
            # default to accepting a JSON response
            headers['Accept'] = 'application/json'

        if self._validate_tokens and auth_method is not None:
            self._auth_mgr.refresh_tokens()

        if auth_method == AuthenticationMethod.SpartanToken:
            headers['x-343-authorization-spartan'] = self._auth_mgr.spartan_token.spartan_token  # type: ignore
        elif auth_method == AuthenticationMethod.ClearanceToken:
            headers['x-343-authorization-spartan'] = self._auth_mgr.spartan_token.spartan_token  # type: ignore
            headers['343-clearance'] = self._auth_mgr.clearance_token.flight_configuration_id  # type: ignore
        elif auth_method == AuthenticationMethod.XSTSv3XboxAudience:
            headers['Authorization'] = self._auth_mgr.xsts_token.authorization_header_value  # type: ignore
            headers['x-xbl-contract-version'] = '3'

        if include_cv:
            headers['MS-CV'] = self._cv.increment()

        return self._auth_mgr.session.request(method, url, headers=headers, **kwargs)

    def get(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('GET', url, **kwargs)

    def options(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('OPTIONS', url, **kwargs)

    def head(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('HEAD', url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('POST', url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('PUT', url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('PATCH', url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> requests.Response:
        return self.request('DELETE', url, **kwargs)

        


""""""

from dataclasses import dataclass

from spnkr.models import CamelModel


@dataclass
class Setting(CamelModel):
    id: str
    value: str


@dataclass
class ProfileUser(CamelModel):
    id: str
    host_id: str
    settings: list[Setting]
    is_sponsored_user: bool


@dataclass
class ProfileResponse(CamelModel):
    profile_users: list[ProfileUser]

    



from dataclasses import dataclass
import datetime as dt
from uuid import UUID

from spnkr.api import enums
from spnkr.models import PascalModel


@dataclass
class MatchCount(PascalModel):
    custom_matches_played_count: int
    matches_played_count: int
    matchmade_matches_played_count: int
    local_matches_played_count: int


@dataclass
class AssetStub(PascalModel):
    asset_kind: enums.AssetKind
    asset_id: UUID
    version_id: UUID


@dataclass
class MatchInfo(PascalModel):
    start_time: dt.datetime
    end_time: dt.datetime
    duration: dt.timedelta
    lifecycle_mode: enums.LifecycleMode
    game_variant_category: enums.GameVariantCategory
    level_id: UUID
    map_variant: AssetStub
    ugc_game_variant: AssetStub
    clearance_id: UUID
    playlist: AssetStub | None
    playlist_experience: enums.PlaylistExperience | None
    playlist_map_mode_pair: AssetStub | None
    season_id: str
    playable_duration: dt.timedelta
    teams_enabled: bool
    team_scoring_enabled: bool


@dataclass
class PlayerMatchHistoryRecord(PascalModel):
    match_id: UUID
    match_info: MatchInfo
    last_team_id: int
    outcome: enums.Outcome
    rank: int
    present_at_end_of_match: bool


@dataclass
class Links(PascalModel):
    pass


@dataclass
class MatchHistoryResponse(PascalModel):
    start: int
    count: int
    result_count: int
    results: list[PlayerMatchHistoryRecord]
    links: Links


@dataclass
class MatchResultPersonalScoreStats(PascalModel):
    name_id: int
    count: int
    total_personal_score_awarded: int


@dataclass
class MatchResultMedalStats(PascalModel):
    name_id: int
    count: int
    total_personal_score_awarded: int


@dataclass
class CoreStats(PascalModel):
    score: int
    personal_score: int
    rounds_won: int
    rounds_lost: int
    rounds_tied: int
    kills: int
    deaths: int
    assists: int
    kda: float
    suicides: int
    betrayals: int
    average_life_duration: dt.timedelta
    grenade_kills: int
    headshot_kills: int
    melee_kills: int
    power_weapon_kills: int
    shots_fired: int
    shots_hit: int
    accuracy: float
    damage_dealt: int
    damage_taken: int
    callout_assists: int
    vehicle_destroys: int
    driver_assists: int
    hijacks: int
    emp_assists: int
    max_killing_spree: int
    medals: list[MatchResultMedalStats]
    personal_scores: list[MatchResultPersonalScoreStats]
    deprecated_damage_dealt: float
    deprecated_damage_taken: float
    spawns: int


@dataclass
class BombStats(PascalModel):
    bomb_carriers_killed: int
    bomb_defusals: int
    bomb_defusers_killed: int
    bomb_detonations: int
    bomb_pick_ups: int
    bomb_plants: int
    bomb_returns: int
    kills_as_bomb_carrier: int
    time_as_bomb_carrier: dt.timedelta


@dataclass
class CaptureTheFlagStats(PascalModel):
    flag_capture_assists: int
    flag_captures: int
    flag_carriers_killed: int
    flag_grabs: int
    flag_returners_killed: int
    flag_returns: int
    flag_secures: int
    flag_steals: int
    kills_as_flag_carrier: int
    kills_as_flag_returner: int
    time_as_flag_carrier: dt.timedelta


@dataclass
class EliminationStats(PascalModel):
    allies_revived: int
    elimination_assists: int
    eliminations: int
    enemy_revives_denied: int
    executions: int
    kills_as_last_player_standing: int
    last_players_standing_killed: int
    rounds_survived: int
    times_revived_by_ally: int
    lives_remaining: int | None
    elimination_order: int


@dataclass
class ExtractionStats(PascalModel):
    extraction_conversions_completed: int
    extraction_conversions_denied: int
    extraction_initiations_completed: int
    extraction_initiations_denied: int
    successful_extractions: int


@dataclass
class InfectionStats(PascalModel):
    pass


@dataclass
class OddballStats(PascalModel):
    kills_as_skull_carrier: int
    longest_time_as_skull_carrier: dt.timedelta
    skull_carriers_killed: int
    skull_grabs: int
    time_as_skull_carrier: dt.timedelta
    skull_scoring_ticks: int


@dataclass
class ZonesStats(PascalModel):
    zone_captures: int
    zone_defensive_kills: int
    zone_offensive_kills: int
    zone_secures: int
    total_zone_occupation_time: dt.timedelta
    zone_scoring_ticks: int


@dataclass
class StockpileStats(PascalModel):
    kills_as_power_seed_carrier: int
    power_seeds_deposited: int
    power_seeds_stolen: int
    power_seed_carriers_killed: int
    time_as_power_seed_carrier: dt.timedelta
    time_as_power_seed_driver: dt.timedelta


@dataclass
class Stats(PascalModel):
    core_stats: CoreStats
    bomb_stats: BombStats | None
    capture_the_flag_stats: CaptureTheFlagStats | None
    elimination_stats: EliminationStats | None
    extraction_stats: ExtractionStats | None
    infection_stats: InfectionStats | None
    oddball_stats: OddballStats | None
    zones_stats: ZonesStats | None
    stockpile_stats: StockpileStats | None


@dataclass
class Team(PascalModel):
    team_id: int
    outcome: enums.Outcome
    rank: int
    stats: Stats


@dataclass
class BotAttributes(PascalModel):
    difficulty: enums.BotDifficulty


@dataclass
class ParticipationInfo(PascalModel):
    first_joined_time: dt.datetime
    last_leave_time: dt.datetime | None
    present_at_beginning: bool
    joined_in_progress: bool
    left_in_progress: bool
    present_at_completion: bool
    time_played: dt.timedelta
    confirmed_participation: bool | None


@dataclass
class PlayerTeamStats(PascalModel):
    team_id: int
    stats: Stats


@dataclass
class Player(PascalModel):
    player_id: str
    player_type: enums.PlayerType
    bot_attributes: BotAttributes | None
    last_team_id: int
    outcome: int
    rank: int
    participation_info: ParticipationInfo
    player_team_stats: list[PlayerTeamStats]


@dataclass
class MatchStats(PascalModel):
    match_id: UUID
    match_info: MatchInfo
    teams: list[Team]
    players: list[Player]




"""Models for the HIUGC_Discovery authority."""

from dataclasses import dataclass
from typing import Any
from uuid import UUID

from spnkr.models import PascalModel, Date


@dataclass
class OnlineUriReference(PascalModel):
    authority_id: str
    path: str
    query_string: str | None
    retry_policy_id: str
    topic_name: str
    acknowledgement_type_id: int
    authentication_lifetime_extension_supported: bool
    clearance_aware: bool


@dataclass
class AssetFiles(PascalModel):
    prefix: str
    file_relative_paths: list[str]
    prefix_endpoint: OnlineUriReference


@dataclass
class AssetStats(PascalModel):
    plays_recent: int
    plays_all_time: int
    favorites: int
    likes: int
    bookmarks: int
    parent_asset_count: int
    average_rating: float | None
    number_of_ratings: int


@dataclass
class Asset(PascalModel):
    asset_id: UUID
    version_id: UUID
    public_name: str
    description: str
    files: AssetFiles
    contributors: list[str]
    asset_home: int
    asset_stats: AssetStats
    inspection_result: int
    clone_behavior: int
    order: int


@dataclass
class ManifestCustomData(PascalModel):
    branch_name: str | None = None
    build_number: str | None = None
    kind: int | None = None
    content_version: str | None = None
    build_guid: UUID | None = None
    visibility: int | None = None


@dataclass
class Manifest(Asset):
    custom_data: ManifestCustomData
    tags: list[str]
    map_links: list[Asset]
    playlist_links: list[Asset]
    ugc_game_variant_links: list[Asset]
    prefab_links: list[Asset] | None = None
    map_mode_pair_links: list[Asset] | None = None
    engine_game_variant_links: list[Asset] | None = None


@dataclass
class Project(Asset):
    custom_data: dict[Any, Any]
    tags: list[str]
    map_links: list[Asset]
    playlist_links: list[Asset]
    prefab_links: list[Asset] | None
    ugc_game_variant_links: list[Asset]
    map_mode_pair_links: list[Asset] | None
    engine_game_variant_links: list[Asset] | None = None


@dataclass
class EngineGameVariantSubsetData(PascalModel):
    stat_bucket_game_type: int
    engine_name: str
    variant_name: str


@dataclass
class EngineGameVariantCustomData(PascalModel):
    subset_data: EngineGameVariantSubsetData
    localized_data: dict[Any, Any]


@dataclass
class EngineGameVariant(Asset):
    custom_data: EngineGameVariantCustomData
    tags: list[str]


@dataclass
class MapCustomData(PascalModel):
    num_of_objects_on_map: int
    tag_level_id: int
    is_baked: bool
    has_node_graph: bool


@dataclass
class Map(Asset):
    custom_data: MapCustomData
    tags: list[str]
    prefab_links: list[Asset] | None


@dataclass
class RotationWeight(PascalModel):
    weight: float


@dataclass
class RotationEntry(Asset):
    metadata: RotationWeight


@dataclass
class PlaylistEntry(PascalModel):
    map_mode_pair_asset_id: str
    metadata: RotationWeight


@dataclass
class PlaylistCustomData(PascalModel):
    playlist_entries: list[PlaylistEntry]
    strategy: int
    min_teams: int
    min_team_size: int
    max_teams: int
    max_team_size: int
    max_team_imbalance: int
    max_splitscreen_players_allowed: int
    allow_friend_join_in_progress: bool
    allow_matchmaking_join_in_progress: bool
    allow_bot_join_in_progress: bool
    exit_experience_duration_sec: int
    fireteam_leader_kick_allowed: bool
    disable_midgame_chat: bool
    allowed_device_inputs: list[int]
    bot_difficulty: int
    min_fireteam_size: int
    max_fireteam_size: int


@dataclass
class Playlist(Asset):
    custom_data: PlaylistCustomData
    tags: list[str]
    rotation_entries: list[RotationEntry]


@dataclass
class MapModePair(Asset):
    custom_data: dict[Any, Any]
    tags: list[str]
    map_link: Asset
    ugc_game_variant_link: Asset


@dataclass
class PrefabCustomData(PascalModel):
    parts: int
    has_node_graph: bool


@dataclass
class Prefab(Asset):
    custom_data: PrefabCustomData
    tags: list[str]


@dataclass
class Tag(PascalModel):
    name: str


@dataclass
class TagsInfo(PascalModel):
    canned_tags: list[Tag]


@dataclass
class UgcGameVariantCustomData(PascalModel):
    key_values: dict[Any, Any]


@dataclass
class UgcGameVariant(Asset):
    custom_data: UgcGameVariantCustomData
    tags: list[str]
    engine_game_variant_link: Asset


@dataclass
class TagCount(PascalModel):
    tag: str
    count: int


@dataclass
class SearchResult(PascalModel):
    asset_id: str
    asset_version_id: str
    name: str
    description: str
    asset_kind: int
    tags: list[str]
    thumbnail_url: str
    referenced_assets: list[Any]  # uncertain
    original_author: str
    likes: int
    bookmarks: int
    plays_recent: int
    number_of_objects: int
    date_created_utc: Date
    date_modified_utc: Date
    date_published_utc: Date
    has_node_graph: bool
    read_only_clones: bool
    plays_all_time: int
    contributors: list[str]
    parent_asset_count: int
    average_rating: float
    number_of_ratings: int


@dataclass
class Search(PascalModel):
    tags: list[TagCount]
    estimated_total: int
    start: int
    count: int
    result_count: int
    results: list[SearchResult]
    links: dict[Any, Any]  # uncertain


@dataclass
class FilmChunk(PascalModel):
    index: int
    chunk_start_time_offset_milliseconds: int
    duration_milliseconds: int
    chunk_size: int
    file_relative_path: str
    chunk_type: int


@dataclass
class FilmCustomData(PascalModel):
    film_length: int
    chunks: list[FilmChunk]
    has_game_ended: bool
    manifest_refresh_seconds: int
    match_id: str
    film_major_version: int


@dataclass
class Film(PascalModel):
    film_status_bond: int
    custom_data: FilmCustomData
    blob_storage_path_prefix: str
    asset_id: UUID





import enum


class AuthenticationMethod(int, enum.Enum):
    SpartanToken = 1
    ClearanceToken = 2
    XSTSv3XboxAudience = 3
    

class ProfileSetting(str, enum.Enum):
    GameDisplayName = "GameDisplayName"
    AppDisplayName = "AppDisplayName"
    AppDisplayPicRaw = "AppDisplayPicRaw"
    GameDisplayPicRaw = "GameDisplayPicRaw"
    PublicGamerpic = "PublicGamerpic"
    ShowUserAsAvatar = "ShowUserAsAvatar"
    Gamerscore = "Gamerscore"
    Gamertag = "Gamertag"
    ModernGamertag = "ModernGamertag"
    ModernGamertagSuffix = "ModernGamertagSuffix"
    UniqueModernGamertag = "UniqueModernGamertag"
    AccountTier = "AccountTier"
    TenureLevel = "TenureLevel"
    XboxOneRep = "XboxOneRep"
    PreferredColor = "PreferredColor"
    Location = "Location"
    Bio = "Bio"
    Watermarks = "Watermarks"
    RealName = "RealName"
    RealNameOverride = "RealNameOverride"
    IsQuarantined = "IsQuarantined"


class ResultOrder(int, enum.Enum):
    Desc = 1
    Asc = 2


class AssetKind(int, enum.Enum):
    Film = 1
    Map = 2
    Playlist = 3
    Prefab = 4
    TestAsset = 5
    UgcGameVariant = 6
    MapModePair = 7
    Project = 8
    Manifest = 9
    EngineGameVariant = 10


class SearchProperty(int, enum.Enum):
    Name = 1
    Likes = 2
    Bookmarks = 3
    PlaysRecent = 4
    NumberOfObjects = 5
    DateCreatedUtc = 6
    DateModifiedUtc = 7
    DatePublishedUtc = 8
    PlaysAllTime = 9
    ParentAssetCount = 10
    AverageRating = 11
    NumberOfRatings = 12


class MatchType(int, enum.Enum):
    All = 1
    Matchmaking = 2
    Custom = 3
    Local = 4


class Outcome(int, enum.Enum):
    Tie = 1
    Win = 2
    Loss = 3
    DidNotFinish = 4


class LifecycleMode(int, enum.Enum):
    Custom = 1
    Matchmade = 3


class GameVariantCategory(int, enum.Enum):
    MultiplayerSlayer = 6
    MultiplayerAttrition = 7
    MultiplayerElimination = 8
    MultiplayerFiesta = 9
    MultiplayerStrongholds = 11
    MultiplayerBastion = 12
    MultiplayerTotalControl = 14
    MultiplayerCtf = 15
    MultiplayerAssault = 16
    MultiplayerExtraction = 17
    MultiplayerOddball = 18
    MultiplayerStockpile = 19
    MultiplayerJuggernaut = 20
    MultiplayerEscort = 23
    MultiplayerGunGame = 24
    MultiplayerGrifball = 25
    TestEngine = 32
    MultiplayerLandGrab = 39


class PlaylistExperience(int, enum.Enum):
    Arena = 2
    BigTeamBattle = 3
    Featured = 5


class PlayerType(int, enum.Enum):
    Human = 1
    Bot = 2


class BotDifficulty(int, enum.Enum):
    # TODO these are guesses
    NA = 0
    Recruit = 1
    Marine = 2
    ODST = 3
    Spartan = 4




"""Endpoints for the "skill" authority."""

from spnkr.api.authorities.base import BaseAuthority
from spnkr.api.authorities.skill import models
from spnkr.api.enums import AuthenticationMethod
from spnkr import util


class SkillAuthority(BaseAuthority):
    URL = 'https://skill.svc.halowaypoint.com:443'

    def get_match_result(self,
                         match_guid: str,
                         player_xuids: list[str]) -> models.MatchSkillInfo:
        """Get player CSR and team MMR values for a given match and player list.

        Args:
            match_guid (str): Halo Infinite match GUID.
            player_xuids (List[str]): The Xbox Live IDs of the match's players.

        Returns:
            MatchResult: The skill data. Top level key is "Value", whose associated
                value is a list of player skill results.
        """

        url = self.URL + f'/hi/matches/{match_guid}/skill'
        params = dict(players=[util.wrap_xuid(x) for x in player_xuids])
        resp = self._session.get(url, auth_method=AuthenticationMethod.ClearanceToken,
                                 params=params)
        resp.raise_for_status()
        return models.MatchSkillInfo.parse_json(resp.text)

    def get_playlist_csr(self,
                         playlist_id: str,
                         player_xuids: list[str]) -> models.PlaylistCsrInfo:
        url = self.URL + f'/hi/playlist/{playlist_id}/csrs'
        params = dict(players=[util.wrap_xuid(x) for x in player_xuids])
        resp = self._session.get(url, auth_method=AuthenticationMethod.ClearanceToken,
                                 params=params)
        resp.raise_for_status()
        return models.PlaylistCsrInfo.parse_json(resp.text)




"""Endpoints for the "halostats" authority."""

from spnkr.api.enums import MatchType
from spnkr.api.authorities.base import BaseAuthority
from spnkr.api.authorities.stats import models
from spnkr import util


class StatsAuthority(BaseAuthority):
    URL = 'https://halostats.svc.halowaypoint.com:443'

    def get_match_count(self, player_xuid: str) -> models.MatchCount:
        """Get match counts across different game experiences for a player.

        The counts returned are for custom matches, matchmade matches, local matches, and
        total matches.

        Args:
            player_xuid (str): Xbox Live ID of the player to get counts for.

        Returns:
            PlayerMatchCount: Match counts.
        """
        xuid = util.wrap_xuid(player_xuid)
        url = self.URL + f'/hi/players/{xuid}/matches/count'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.MatchCount.parse_json(resp.text)

    def get_match_history(self,
                          player_xuid: str,
                          start: int = 0,
                          count: int = 25,
                          match_type: MatchType = MatchType.All) -> models.MatchHistoryResponse:
        """Request a batch of matches from a player's match history.

        Args:
            player_xuid (str): Xbox Live ID of the player to request matches for.
            start (int, optional): The number of matches to skip (offset). Starts at 0.
            count (int, optional): The number of matches to request. Max is 25.
                Cannot exceed the default value. The service will still return a valid
                response with only the default match count, but the upper limit is
                enforced here to ensure that the offset isn't incremented incorrectly
                when attempting to collect continuous match data for a player.
            match_type (enums.MatchType): The type of matches to return. Defaults to "All".

        Returns:
            MatchHistory: _description_
        """
        xuid = util.wrap_xuid(player_xuid)
        url = self.URL + f'/hi/players/{xuid}/matches'
        params = {
            'start': start,
            'count': count,
            'type': match_type.name,
        }
        resp = self._session.get(url, params=params)
        resp.raise_for_status()
        return models.MatchHistoryResponse.parse_json(resp.text)

    def get_match_stats(self, match_guid: str) -> models.MatchStats:
        """Request match details using the Halo Infinite match GUID.

        Args:
            match_guid (str): Halo Infinite GUID identifying the match.

        Returns:
            MatchStats: Match details.
        """
        url = self.URL + f'/hi/matches/{match_guid}/stats'
        resp = self._session.get(url)
        resp.raise_for_status()
        return models.MatchStats.parse_json(resp.text)

